"""A Python3 implementation of the Ant Colony Optimization Meta-Heuristic.

.. moduleauthor:: Robert Grant <rhgrant10@gmail.com>

"""

from .ant import Ant 
from .world import World, Edge, Node
from .solver import Solver
